import json
import logging
import asyncio
import dns.asyncresolver
from domain_email_records.logger import Logger
from domain_email_records import __title__


logger = logging.getLogger(__title__)


class DomainEmailRecordsException(Exception):
    pass


class DomainEmailRecords:

    chunk_size: int
    csv_column: int
    max_query_seconds: int

    def __init__(self, chunk_size=100, csv_column=2, max_query_seconds=5, logging_level="INFO"):
        self.chunk_size = chunk_size
        self.csv_column = csv_column
        self.max_query_seconds = max_query_seconds

        global logger
        logger = Logger(name=__title__).setup(level=logging_level)

    async def process(self, filename: str = None, domains: list = None):
        logger.debug(f"process(filename={filename}, domains={domains})")

        if filename and domains:
            raise DomainEmailRecordsException("Must supply either --filename or --domains not both")

        if filename:
            domains = self.read_domains_file(filename=filename, csv_column=self.csv_column)

        def list_chunk(seq, size):
            return (seq[pos : pos + size] for pos in range(0, len(seq), size))

        for chunk in list_chunk(domains, self.chunk_size):
            results = await asyncio.gather(*[self.domain_record_lookups(domain) for domain in chunk])
            for result in results:
                print(json.dumps(result, indent="  "))

    async def domain_record_lookups(self, domain_name: str) -> dict:
        """
        Query for mx, spf (via apex-TXT) and dmarc records

        :param domain_name:
        :return:
        """
        logger.debug(f"dns_lookups(domain_name={domain_name})")

        records = {}

        # mx
        answers = await self.dns_query(domain_name, "mx")
        if answers:
            for rdata in answers:
                if "mx" not in records:
                    records["mx"] = []
                records["mx"].append(str(rdata.exchange))

        # spf
        answers = await self.dns_query(domain_name, "txt")
        if answers:
            for rdata in answers:
                txt_record = rdata.strings[0].decode("utf8").lower()
                if "spf1" in txt_record:
                    if "spf" not in records:
                        records["spf"] = []
                    records["spf"].append(rdata.strings[0].decode("utf8"))

        # dmarc
        answers = await self.dns_query(f"_dmarc.{domain_name}", "txt")
        if answers:
            for rdata in answers:
                if "dmarc" not in records:
                    records["dmarc"] = []
                records["dmarc"].append(rdata.strings[0].decode("utf8"))

        return {domain_name: records}

    async def dns_query(self, domain_name, query_type):
        try:
            answers = await dns.asyncresolver.resolve(domain_name, query_type.upper(), lifetime=self.max_query_seconds)
        except dns.exception.DNSException:
            answers = None
        return answers

    def read_domains_file(self, filename: str, csv_column: int = 2) -> list:
        """
        Loads domain names from a plain file with domain names only; or detects the comma(,)
        character, then splits and takes the csv_column_index

        :param filename:
        :param csv_column:
        :return:
        """

        domains = []

        with open(filename, "r") as f:
            line = f.readline()
            while line:
                if "," not in line:
                    domains.append(line.strip())
                else:
                    columns = line.strip().split(",")
                    if len(columns) >= csv_column:
                        domains.append(columns[csv_column - 1].strip())
                line = f.readline()

        return domains



import sys
import json
import asyncio
import argparse
import textwrap
from signal import signal, SIGINT
from domain_email_records import __title__
from domain_email_records import __version__
from domain_email_records.domain_email_records import DomainEmailRecords
from domain_email_records.domain_email_records import DomainEmailRecordsException


def sigint_handler(__signal_received, __frame):
    print("SIGINT received, exiting.")
    exit(1)


def main():
    signal(SIGINT, sigint_handler)
    args = __argparse()

    if args.quiet:
        logging_level = "CRITICAL"
    elif args.verbose:
        logging_level = "DEBUG"
    else:
        logging_level = "INFO"

    try:
        asyncio.run(
            DomainEmailRecords(
                chunk_size=args.chunk_size,
                csv_column=args.csv_column,
                logging_level=logging_level,
            ).process(filename=args.filename, domains=args.domains)
        )
    except DomainEmailRecordsException as e:
        error = {"program": "{}".format(__title__), "version": "v{}".format(__version__), "message": str(e.args[0])}
        if len(e.args) > 1:
            error["data"] = str(e.args[1 : len(e.args)])

        print(json.dumps({"error": error}, indent="  "))
        exit(1)


def __argparse() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        add_help=True,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """\
            {name} v{version}

            CLI tool to quickly lookup MX, SPF, DMARC records for many domains
        """.format(
                name=__title__, version=__version__
            )
        ),
    )

    output_parser = parser.add_mutually_exclusive_group(required=False)
    output_parser.add_argument("-q", "--quiet", action="store_true", help="Set quiet logging output")
    output_parser.add_argument("-v", "--verbose", action="store_true", help="Set verbose logging output")

    parser.add_argument(
        "--chunk-size",
        metavar="<size>",
        required=False,
        type=int,
        default=100,
        help="Number of domain names to async resolve together (default: 100)",
    )

    direct_list_parser = parser.add_argument_group(title="direct")
    direct_list_parser.add_argument(
        "-d", "--domains", metavar="<domain>", nargs="*", help="Space separated list of domain names to query"
    )

    file_list_parser = parser.add_argument_group(title="filename")
    file_list_parser.add_argument(
        "-f",
        "--filename",
        metavar="<filename>",
        type=str,
        help="Filename with list of domains to use; either a plain text file list -or- a comma-separated CSV file list.",
    )
    file_list_parser.add_argument(
        "--csv-column",
        metavar="<col>",
        required=False,
        type=int,
        default=2,
        help="CSV column number to use for domain-names -if- the file is CSV formatted (default: 2)",
    )

    args = parser.parse_args()

    if len(sys.argv) < 2 or (args.filename and args.domains):
        parser.print_help()
        print()
        exit(1)

    return args



import time
import logging


class LoggerException(Exception):
    pass


class Logger:

    name = None
    logging = None
    default_level = "info"

    def __init__(self, name):
        self.name = name
        self.logging = logging.getLogger(self.name)

    def setup(self, level=None, fmt="%(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%dT%H:%M:%S%z"):

        logger_init = logging.getLogger(self.name)
        stream_handler = logging.StreamHandler()

        if level is not None:
            log_level = level.upper()
        else:
            log_level = self.default_level.upper()

        if log_level == "CRITICAL":
            logger_init.setLevel(logging.CRITICAL)
            stream_handler.setLevel(logging.CRITICAL)
        elif log_level == "ERROR":
            logger_init.setLevel(logging.ERROR)
            stream_handler.setLevel(logging.ERROR)
        elif log_level in ("WARNING", "WARN"):
            logger_init.setLevel(logging.WARNING)
            stream_handler.setLevel(logging.WARNING)
        elif log_level == "INFO":
            logger_init.setLevel(logging.INFO)
            stream_handler.setLevel(logging.INFO)
        elif log_level == "DEBUG":
            logger_init.setLevel(logging.DEBUG)
            stream_handler.setLevel(logging.DEBUG)
        else:
            raise LoggerException("unknown logger level value", log_level)

        logging.Formatter.converter = time.localtime
        stream_handler.setFormatter(LoggerColorFormatter(fmt=fmt, datefmt=datefmt))
        logger_init.addHandler(stream_handler)

        self.logging = logger_init
        return self.logging


class LoggerColorFormatter(logging.Formatter):

    color_line = "\x1b[90m"  # grey
    color_reset = "\x1b[0m"  # reset

    def __init__(self, **kwargs):
        if "fmt" in kwargs:
            kwargs["fmt"] = "{}{}{}".format(self.color_line, kwargs["fmt"], self.color_reset)
        logging.Formatter.__init__(self, **kwargs)

    def format(self, record):

        levelname = record.levelname.upper()

        if levelname == "CRITICAL":
            color_code = "\x1b[41m"  # white-on-red
        elif levelname == "ERROR":
            color_code = "\x1b[31m"  # red
        elif levelname in ("WARNING", "WARN"):
            color_code = "\x1b[33m"  # yellow
        elif levelname == "INFO":
            color_code = "\x1b[36m"  # cyan
        elif levelname == "DEBUG":
            color_code = "\x1b[37m"  # white
        else:
            color_code = "\x1b[90m"  # grey

        record.levelname = "{}{}{}{}".format(color_code, levelname, self.color_reset, self.color_line)
        return logging.Formatter.format(self, record)
